/* Copyright (C) 2006  Andreas Kunz
 * 
 * YAReG is free software; you can redistribute and/or modify it under the terms of the 
 * GNU General Public License (version 2 or later) as published by the Free Software Foundation.
 * This program is distributed WITHOUT ANY WARRANTY; See the GNU GPL for more details. 
 */

using System;
using System.Collections;
using System.Windows.Forms;

namespace YAReG
{
	/// <summary>
	/// Class capable of "comparing" two dirList entries according to one of the attributes 
	/// (name, size, date, attribute string). Used by the ListView for sorting.
	/// </summary>
	public class DirListComparer : IComparer
	{
		private int columnToSort = 0;
		private SortOrder order = SortOrder.Ascending; // default

		public int ColumnToSort
		{
			get 
			{
				return columnToSort; 
			}

			set 
			{
				columnToSort = value; 
			}
		}

		public SortOrder Order
		{
			get 
			{
				return order; 
			}

			set 
			{
				order = value; 
			}
		}



		/// <summary>
		/// Compares two FsListItems, used for sorting the dirList.
		/// 
		/// The result of the comparison: 
		/// "0" if equal, negative if 'firstItem' is smaller than 'secondItem' 
		/// and positive if 'firstItem' is greater than 'secondItem'.
		/// </summary>
		/// <param name="firstItem"></param>
		/// <param name="secondItem"></param>
		/// <returns></returns>
		public int Compare(object firstItem, object secondItem) 
		{
			// Cast the objects to fsListItem
			FsListItem first = (FsListItem) firstItem;
			FsListItem second = (FsListItem) secondItem;

			// determine whether to invert the "regular" result depending on sort order
			bool invert = (order == SortOrder.Descending);

			// call column-specific comparer
			switch (columnToSort) 
			{
				// "Name" column
				case 0:
					return compareByName(first, second, invert);

				// "Size" column
				case 1:
					return compareBySize(first, second, invert);

				// "Modified" column
				case 2:
					return compareByDate(first, second, invert);

				// "Attributes" column
				case 3:
					return compareByAttributes(first, second, invert);

				default:
					return 0;
			}
		}


		/// <summary>
		/// Compare the given FsListItems by name
		/// (default: ascending order, descending if invert == true)
		/// </summary>
		/// <param name="first"></param>
		/// <param name="second"></param>
		/// <param name="invert"></param>
		/// <returns></returns>
		private int compareByName(FsListItem first, FsListItem second, bool invert) 
		{
			int result;

			// folders are always "smaller" than files
			if (first.Node.NodeType == FsNode.fsType.DIRECTORY
				&& second.Node.NodeType != FsNode.fsType.DIRECTORY) 
			{
				result = -1;
			} 
			else 
			{
				if (second.Node.NodeType == FsNode.fsType.DIRECTORY
					&& first.Node.NodeType != FsNode.fsType.DIRECTORY) 
				{
					result = 1;
				} 
				else 
				{
					// same type -> sort regularly
					result = String.Compare(first.Node.Text, second.Node.Text);
				}
			}

			// comparison was according to ascending sort order; invert if desired
			return (invert ? (- result) : result);
		}


		/// <summary>
		/// Compare the given FsListItems by size 
		/// (default: ascending order, descending if invert == true)
		/// </summary>
		/// <param name="first"></param>
		/// <param name="second"></param>
		/// <param name="invert"></param>
		/// <returns></returns>
		private int compareBySize(FsListItem first, FsListItem second, bool invert) 
		{
			long firstSize = first.Node.Size;
			long secondSize = second.Node.Size;
			int result;

			// filesizes can be larger than int! -> return -1/0/+1
			// folders have size==-1, so they are sorted correctly
			result = Math.Sign(first.Node.Size - second.Node.Size);

			// but sort items with identical size (including folders) by name
			if (result == 0)
			{
				result = String.Compare(first.Node.Text, second.Node.Text);
			}

			return (invert ? (- result) : result);
		}


		/// <summary>
		/// Compare the given FsListItems by date 
		/// (default: ascending order, descending if invert == true)
		/// </summary>
		/// <param name="first"></param>
		/// <param name="second"></param>
		/// <param name="invert"></param>
		/// <returns></returns>
		private int compareByDate(FsListItem first, FsListItem second, bool invert) 
		{
			DateTime firstDate = first.Node.DateTime;
			DateTime secondDate = second.Node.DateTime;
			int result;

			// folders are always "smaller" than files
			if (first.Node.NodeType == FsNode.fsType.DIRECTORY
				&& second.Node.NodeType != FsNode.fsType.DIRECTORY) 
			{
				result = -1;
			} 
			else 
			{
				if (second.Node.NodeType == FsNode.fsType.DIRECTORY
					&& first.Node.NodeType != FsNode.fsType.DIRECTORY) 
				{
					result = 1;
				} 
				else 
				{
					// same type -> sort regularly
					result = DateTime.Compare(firstDate, secondDate);
				}
			}

			return (invert ? (- result) : result);
		}


		/// <summary>
		/// Compare the given FsListItems by attribute string 
		/// (default: ascending order, descending if invert == true)
		/// </summary>
		/// <param name="first"></param>
		/// <param name="second"></param>
		/// <param name="invert"></param>
		/// <returns></returns>
		private int compareByAttributes(FsListItem first, FsListItem second, bool invert) 
		{
			string firstAttributes = first.Node.Attributes;
			string secondAttributes = second.Node.Attributes;
			int result;

			// folders are always "smaller" than files
			if (first.Node.NodeType == FsNode.fsType.DIRECTORY
				&& second.Node.NodeType != FsNode.fsType.DIRECTORY) 
			{
				result = -1;
			} 
			else 
			{
				if (second.Node.NodeType == FsNode.fsType.DIRECTORY
					&& first.Node.NodeType != FsNode.fsType.DIRECTORY) 
				{
					result = 1;
				} 
				else 
				{
					// same type -> sort regularly
					result = String.Compare(first.Node.Attributes, second.Node.Attributes);

					// but sort items with identical attributes by name
					if (result == 0)
					{
						result = String.Compare(first.Node.Text, second.Node.Text);
					}
				}
			}

			return (invert ? (- result) : result);
		}
	}
}
