/* Copyright (C) 2004, 2005, 2006  Andreas Kunz
 * 
 * YAReG is free software; you can redistribute and/or modify it under the terms of the 
 * GNU General Public License (version 2 or later) as published by the Free Software Foundation.
 * This program is distributed WITHOUT ANY WARRANTY; See the GNU GPL for more details. 
 */

using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace YAReG
{
	/// <summary>
	/// A File System Node
	/// </summary>
	public class FsNode : TreeNode
	{
		// protected members
		protected string		name;
		protected string		path = "";
		protected PartitionNode partition;
		protected long			size = 0;
		protected string		sizeString = "";
		protected string		date;
		protected DateTime		dateTime;
		protected string		attributes;
		protected fsType		nodeType;
		protected bool			isPopulated = false;
		protected ArrayList		files;

		#region public accessors
		public string Name
		{
			get 
			{
				return name; 
			}
		}

		public string Path
		{
			get 
			{
				return path; 
			}
		}

		public PartitionNode Partition 
		{
			get 
			{
				return partition; 
			}
		}

		public long Size 
		{
			get 
			{
				return size; 
			}
		}

		public string SizeString
		{
			get 
			{
				return sizeString; 
			}
		}

		public string Date
		{
			get
			{
				return date; 
			}
		}

		public DateTime DateTime
		{
			get 
			{
				return dateTime; 
			}
		}
		public string Attributes
		{
			get 
			{
				return attributes; 
			}
		}

		public fsType NodeType
		{
			get 
			{
				return nodeType; 
			}
		}

		public bool IsPopulated
		{
			get 
			{
				return isPopulated; 
			}
		}

		public ArrayList Files
		{
			get 
			{
				return files; 
			}
		}
		#endregion


		/// <summary>
		/// Empty constructor for dummy nodes and as base constructor for 
		/// specialized class PartitionNode.
		/// </summary>
		protected FsNode() 
		{
			// do nothing
		}


		/// <summary>
		/// Creates a fsNode (file or directory) from raw command line tool output
		/// </summary>
		/// <param name="line">raw data from a command line tool</param>
		/// <param name="source">identifier of the tool</param>
		/// <param name="part">partition</param>
		public FsNode(string line, From source, PartitionNode part, string path)
		{
			switch (source) 
			{
				case From.RFSTOOL_FILE:
					name		= Regex.Replace(line, @".*:\d\d\s\d\d\d\d\s(.*)$", "$1");
					name		= Name.Trim();
					this.path	= path;
					partition	= part;
					dateTime	= convertDate(line.Substring(28, 20));
					date		= dateTime.ToShortDateString() + " " + dateTime.ToShortTimeString();
					attributes	= line.Substring(0, 10);
					nodeType	= GetTypeFromNameAndAttributes(name, attributes);
					if (nodeType != fsType.DIRECTORY) 
					{
						sizeString	= line.Substring(11, 12);
						size		= Int64.Parse(sizeString);
						sizeString	= buildSizeString(Size);
					} 
					else 
					{
						SelectedImageIndex = 2;
						// append dummy node
						this.Nodes.Add(CreateDummy()); 
						size = -1; // smaller than all files; obviously no actual value
					}
					break;

				default:
					MessageBox.Show("Commandline tool data source " + 
						source + " unknown.", "Error");
					break;
			}

			Text = name;
			ImageIndex = (int) nodeType;
			files = new ArrayList();
		}


		/// <summary>
		/// Dummy nodes are used for unpopulated nodes that can have children 
		/// (directories and partitions).
		/// </summary>
		/// <returns></returns>
		protected static FsNode CreateDummy()
		{
			FsNode dummy = new FsNode();
			dummy.nodeType = fsType.DUMMY;
			return dummy;
		}


		/// <summary>
		/// Guesses the file type from a file's name and its Attributes
		/// </summary>
		/// <param name="filename">Name of the file</param>
		/// <param name="Attributes">Attributes string</param>
		/// <returns>fsType</returns>
		private fsType GetTypeFromNameAndAttributes(string filename, string attributes)
		{
			// directories are easy to detect
			if (attributes.Substring(0,1) == "d")
				return fsType.DIRECTORY;

			// assume files marked as "executable" for somebody are some kind of executables
			if (attributes.IndexOf("x") > 0)
			{ 
				return fsType.EXECUTABLE;
			}

			// for the remaining files do extension guessing
			// TODO(enhancement): this heuristics could be extended if it seems appropriate
			string extension = filename.Substring(filename.LastIndexOf(".") + 1);
			switch (extension)
			{
				case "so":
					return fsType.LIBRARY;
				default:
					return fsType.ANYFILE;
			}
		}


		/// <summary>
		/// Populates this node, i.e. calls rfstool to read the contents of this directory 
		/// or partition.
		/// </summary>
		internal void Populate()
		{
			// plausibility check: only allow populating partitions and directories
			if (nodeType != fsType.DIRECTORY
				&& nodeType != fsType.PARTITION) 
			{
				// TODO: log this problem... once logging is available...
				return;
			}

			Nodes.Clear(); // remove dummy child
			string childrenPath;
			if (nodeType == fsType.DIRECTORY)
				childrenPath = path + "/" + name;
			else // partition
				childrenPath = "";

			Process p = new Process();
			p.StartInfo.UseShellExecute = false;
			p.StartInfo.RedirectStandardOutput = true;
			p.StartInfo.FileName = YAReGForm.RFSTOOL;
			string arguments = "-p" + Partition.PartID + " " + "\"" + childrenPath + "\"";
			p.StartInfo.Arguments = arguments;
			p.StartInfo.CreateNoWindow = true;
			try 
			{
				p.Start();
				StreamReader output = p.StandardOutput;
				ArrayList fileNodes = new ArrayList();
				FsNode newChild;

				string line;
				while (output.Peek() > -1) 
				{
					line = output.ReadLine();
					if (Regex.IsMatch(line, "^[d-][r-].*")) 
					{
						newChild = new FsNode(line, From.RFSTOOL_FILE, 
							this.Partition,	childrenPath);
						if(newChild.nodeType == FsNode.fsType.DIRECTORY)
							this.Nodes.Add(newChild);
						else
							this.files.Add(newChild);
					}
				}
				p.WaitForExit();
			}
			catch (Exception e) 
			{
				MessageBox.Show("An exception occurred while calling rfstool to get directory data from directory \"" +
					childrenPath + "\": "
					+ e.Message + "\nPlease inform the YAReG author about this problem. The arguments passed to rfstool are: " + arguments, 
					"Error");
			}

			// remove "." and ".." directories
			// TODO: check whether it makes sense to display ".." to navigate to the parent folder
			try 
			{
				this.Nodes.RemoveAt(0);
				this.Nodes.RemoveAt(0); 
			} 
			catch
			{
				MessageBox.Show("An exception occurred when removing the . and .. nodes " +
					"from the directory content. \nThis means something has already gone " +
					"wrong before, probably rfstool has encountered an error. \nNot much " +
					"help I can offer on that... just try to ignore this message...  " +
					"\nrfstool.exe was previously called with the following arguments: " + arguments, "Error");
			}

			isPopulated = true;
		}


		/// <summary>
		/// Parses the date and time given by rfstool and creates a DateTime object from it.
		/// </summary>
		/// <param name="dateString"></param>
		/// <returns></returns>
		private DateTime convertDate(string dateString) 
		{
			// dateString is like "Jan 01 01:00:00 1970"	
			int year = Int32.Parse(dateString.Substring(16));
			int month = monthToInt(dateString.Substring(0, 3));
			int day = Int32.Parse(dateString.Substring(4, 2));
			int hour = Int32.Parse(dateString.Substring(7, 2));
			int minute = Int32.Parse(dateString.Substring(10, 2));
			int second = Int32.Parse(dateString.Substring(13, 2));

			DateTime dateTime = new DateTime(year, month, day, hour, minute, second);
			return dateTime;
		}


		/// <summary>
		/// Parses a three-letter abbreviation of a month and returns the month's number (1-12).
		/// </summary>
		/// <param name="shortMonth"></param>
		/// <returns></returns>
		private int monthToInt(string shortMonth) 
		{
			switch(shortMonth) 
			{
				case "Jan":
					return 1;

				case "Feb":
					return 2;

				case "Mar":
					return 3;

				case "Apr":
					return 4;

				case "May":
					return 5;

				case "Jun":
					return 6;

				case "Jul":
					return 7;

				case "Aug":
					return 8;

				case "Sep":
					return 9;

				case "Oct":
					return 10;

				case "Nov":
					return 11;

				case "Dec":
					return 12;

				default:
					MessageBox.Show("Error when parsing month \"" + shortMonth + "\"");
					return 1; // this is wrong, but crashing is worse than a wrong date
			}
		}


		/// <summary>
		/// Converts the given number of bytes to kilobytes and returns as string.
		/// </summary>
		/// <param name="bytes"></param>
		/// <returns></returns>
		private string buildSizeString(long bytes) 
		{
			if (bytes > 0) 
			{
				bytes += 512;
				bytes /= 1024;
				if (bytes == 0)
					bytes += 1;
			}
			return bytes.ToString() + " KB";
		}


		/// <summary>
		/// The source of raw commandline tool output data to be parsed
		/// </summary>
		public enum From
		{
			RFSTOOL_FILE,
			RFSTOOL_PART
		}


		/// <summary>
		/// The different types of fsNodes.
		/// The numbers correspond to ImageIndex in imlSmallIcons.
		/// </summary>
		public enum fsType
		{
			PARTITION = 0,
			DIRECTORY = 1,
			DIRECTORY_OPEN = 2, // no real type, only for ImageIndex
			ANYFILE = 3,
			EXECUTABLE = 4,
			LIBRARY = 5,
			TEXT = 6,
			BATCHFILE = 7,
			SYSCONFIG = 8,
			DUMMY = 666        // dummy nodes are needed for unpopulated nodes
			// have a nice "+" sign to expand them
		}
	}
}
